from nn_rag import Controller
import os
import warnings

warnings.simplefilter(action='ignore', category=FutureWarning)
warnings.simplefilter(action='ignore', category=DeprecationWarning)

__author__ = 'Darryl Oatridge'


def run_hadron_controller():
    # get the controller run attributes from Environment
    uri_pm_repo = os.environ['HADRON_DOMAIN_REPO_PATH']
    if not uri_pm_repo:
        raise KeyError("The mandatory environment variable 'HADRON_DOMAIN_REPO_PATH' has not been set")

    run_book = None if not os.getenv('HADRON_CONTROLLER_RUNBOOK') else os.getenv('HADRON_CONTROLLER_RUNBOOK')
    repeat = None if not os.getenv('HADRON_CONTROLLER_REPEAT') else os.getenv('HADRON_CONTROLLER_REPEAT')
    sleep = None if not os.getenv('HADRON_CONTROLLER_SLEEP') else os.getenv('HADRON_CONTROLLER_SLEEP')
    run_time = None if not os.getenv('HADRON_CONTROLLER_RUNTIME') else os.getenv('HADRON_CONTROLLER_RUNTIME')
    run_cycle_report = None if not os.getenv('HADRON_CONTROLLER_REPORT') else os.getenv('HADRON_CONTROLLER_REPORT')
    source_check_uri = None if not os.getenv('HADRON_CONTROLLER_SOURCE_CHECK') else os.getenv(
        'HADRON_CONTROLLER_SOURCE_CHECK')

    # instantiate the Controller passing any remaining kwargs
    controller = Controller.from_env(uri_pm_repo=uri_pm_repo, default_save=False, has_contract=True)
    # run the arrows nano services.
    controller.run_controller(run_book=run_book, repeat=repeat, sleep=sleep, run_time=run_time,
                              source_check_uri=source_check_uri, run_cycle_report=run_cycle_report)


def set_env_from_payload_if_available():
    payload = {
        "hadron_kwargs":{
            "HADRON_DOMAIN_REPO_PATH": "https://raw.githubusercontent.com/project-hadron/neural_rag/main/jupyter/knowledge/hadron/contracts/",
            "HADRON_KNOWLEDGE_SOURCE_URI": "https://www.europarl.europa.eu/doceo/document/TA-9-2024-0138_EN.pdf",
            "HADRON_KNOWLEDGE_PERSIST_URI": "./cache/",
            "HADRON_PROFILE_NAME": "EU_AI_Act_2024",
        }
    }

    # extract any extra kwargs
    hadron_kwargs = payload.get('hadron_kwargs', {})
    # export and pop any environment variable from the kwargs
    for key in tuple(hadron_kwargs.keys()):
        key = key.upper()
        os.environ[key] = hadron_kwargs.pop(key)

if __name__ == '__main__':
    set_env_from_payload_if_available()
    run_hadron_controller()



import unittest
import os
from pathlib import Path
import shutil
from datetime import datetime
import pandas as pd
import pyarrow as pa
import pyarrow.compute as pc
from ds_core.properties.property_manager import PropertyManager
from nn_rag.components.commons import Commons
from nn_rag import Knowledge
from nn_rag.intent.knowledge_intent import KnowledgeIntent


# Pandas setup
pd.set_option('max_colwidth', 320)
pd.set_option('display.max_rows', 100)
pd.set_option('display.max_columns', 99)
pd.set_option('expand_frame_repr', True)


class KnowledgeIntentTest(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        pass

    @classmethod
    def tearDownClass(cls):
        pass

    def setUp(self):
        # clean out any old environments
        for key in os.environ.keys():
            if key.startswith('HADRON'):
                del os.environ[key]
        # Local Domain Contract
        os.environ['HADRON_PM_PATH'] = os.path.join('working', 'contracts')
        os.environ['HADRON_PM_TYPE'] = 'parquet'
        # Local Connectivity
        os.environ['HADRON_DEFAULT_PATH'] = Path('working/data').as_posix()
        # Specialist Component
        try:
            os.makedirs(os.environ['HADRON_PM_PATH'])
        except OSError:
            pass
        try:
            os.makedirs(os.environ['HADRON_DEFAULT_PATH'])
        except OSError:
            pass
        try:
            shutil.copytree('../_test_data', os.path.join(os.environ['PWD'], 'working/source'))
        except OSError:
            pass
        PropertyManager._remove_all()

    def tearDown(self):
        try:
            shutil.rmtree('working')
        except OSError:
            pass

    def test_text_join(self):
        kn = Knowledge.from_memory()
        tools: KnowledgeIntent = kn.tools
        text = ('You took too long. You are not easy to deal with. Payment Failure/Incorrect Payment.\n\nYou provided '
                'me with incorrect information. Unhappy with delay.\n\nUnsuitable advice. You never answered my question.'
                'You did not understand my needs.\n\nI have been mis-sold. My details are not accurate.')
        arr = pa.array([text], pa.string())
        tbl = pa.table([arr], names=['text'])
        result = tools.text_to_paragraphs(tbl, has_stats=False)
        print(kn.table_report(result, head=5).to_string())
        result = tools.text_join(result)
        print(kn.table_report(result, head=5).to_string())

    def test_str_remove_text_index(self):
        kn = Knowledge.from_memory()
        tools: KnowledgeIntent = kn.tools
        text = ('You took too long. You are not easy to deal with. Payment Failure/Incorrect Payment.\n\nYou provided '
                'me with incorrect information. Unhappy with delay.\n\nUnsuitable advice. You never answered my question.'
                'You did not understand my needs.\n\nI have been mis-sold. My details are not accurate.')
        arr = pa.array([text], pa.string())
        tbl = pa.table([arr], names=['text'])
        result = tools.text_to_paragraphs(tbl)
        print(kn.table_report(result, head=5).to_string())
        result = tools.filter_on_mask(result, indices=[0, (2, 7)])
        print(kn.table_report(result, head=5).to_string())

    def test_str_remove_text_pattern(self):
        kn = Knowledge.from_memory()
        tools: KnowledgeIntent = kn.tools
        text = ('You took too long. You are not easy to deal with. Payment Failure/Incorrect Payment.\n\nYou provided '
                'me with incorrect information. Unhappy with delay.\n\nUnsuitable advice. You never answered my question.'
                'You did not understand my needs.\n\nI have been mis-sold. My details are not accurate.')
        arr = pa.array([text], pa.string())
        tbl = pa.table([arr], names=['text'])
        result = tools.text_to_paragraphs(tbl)
        print(kn.table_report(result, head=5).to_string())
        result = tools.filter_on_mask(result, pattern='^You.*(You|Unhappy)')
        print(kn.table_report(result, head=5).to_string())


    def test_text_to_paragraph(self):
        kn = Knowledge.from_memory()
        tools: KnowledgeIntent = kn.tools
        uri = "https://assets.circle.so/kvx4ix1f5ctctk55daheobna46hf"
        tbl = kn.set_source_uri(uri).load_source_canonical()
        # text = ('You took too long. You are not easy to deal with. Payment Failure/Incorrect Payment. You provided '
        #         'me with incorrect information. Unhappy with delay. Unsuitable advice. You never answered my question.\n\n'
        #         'You did not understand my needs. I have been mis-sold. My details are not accurate. You have asked '
        #         'for too much information. You were not helpful. Payment not generated/received by customer. You did '
        #         'not keep me updated. Incorrect information given. The performance of my product was poor.\n\n No reply '
        #         'to customer contact. Requested documentation not issued. You did not explain the terms & conditions.\n\n'
        #         'Policy amendments not carried out. You did not explain the next steps/process to me. I cannot '
        #         'understand your letter/comms. Standard letter inappropriate. Customer payment processed incorrectly.\n\n'
        #         'All points not addressed. Could not understand the agent. Issue with terms and conditions. Misleading '
        #         'information. I can not use the customer portal. your customer portal is unhelpful.')
        # arr = pa.array([text], pa.string())
        # tbl = pa.table([arr], names=['text'])
        result = tools.text_to_paragraphs(tbl)
        print(kn.table_report(result, head=5).to_string())

    def test_text_to_sentence(self):
        kn = Knowledge.from_memory()
        tools: KnowledgeIntent = kn.tools
        text = ('You took too long. You are not easy to deal with. Payment Failure/Incorrect Payment. You provided '
                'me with incorrect information. Unhappy with delay. Unsuitable advice. You never answered my question. '
                'You did not understand my needs. I have been mis-sold. My details are not accurate. You have asked '
                'for too much information. You were not helpful. Payment not generated/received by customer. You did '
                'not keep me updated. Incorrect information given. The performance of my product was poor. No reply '
                'to customer contact. Requested documentation not issued. You did not explain the terms & conditions. '
                'Policy amendments not carried out. You did not explain the next steps/process to me. I cannot '
                'understand your letter/comms. Standard letter inappropriate. Customer payment processed incorrectly. '
                'All points not addressed. Could not understand the agent. Issue with terms and conditions. Misleading '
                'information. I can not use the customer portal. your customer portal is unhelpful')
        arr = pa.array([text], pa.string())
        tbl = pa.table([arr], names=['text'])
        result = tools.text_to_sentences(tbl)
        print(kn.table_report(result, head=5).to_string())
        result = tools.text_to_sentences(tbl, has_stats=False)
        print(kn.table_report(result, head=5).to_string())

    def test_text_to_sentence_max(self):
        kn = Knowledge.from_env('tester', has_contract=False)
        tools: KnowledgeIntent = kn.tools
        text = ('You took too long. You are not easy to deal with. Payment Failure/Incorrect Payment. You provided '
                'me with incorrect information. Unhappy with delay. Unsuitable advice. You never answered my question. '
                'You did not understand my needs. I have been mis-sold. My details are not accurate. You have asked '
                'for too much information. You were not helpful. Payment not generated/received by customer. You did '
                'not keep me updated. Incorrect information given. The performance of my product was poor. No reply '
                'to customer contact. Requested documentation not issued. You did not explain the terms & conditions. '
                'Policy amendments not carried out. You did not explain the next steps/process to me. I cannot '
                'understand your letter/comms. Standard letter inappropriate. Customer payment processed incorrectly. '
                'All points not addressed. Could not understand the agent. Issue with terms and conditions. Misleading '
                'information. I can not use the customer portal. your customer portal is unhelpful')
        arr = pa.array([text], pa.string())
        tbl = pa.table([arr], names=['text'])
        # uri = "https://pressbooks.oer.hawaii.edu/humannutrition2/open/download?type=pdf"
        # tbl = kn.set_source_uri(uri, file_type='pdf').load_source_canonical()
        result =  tools.text_to_sentences(tbl, max_char_size=100)
        tprint(result, headers=['sentence_score', 'char_count', 'word_count'])


    def test_text_chunk(self):
        kn = Knowledge.from_memory()
        tools: KnowledgeIntent = kn.tools
        uri = "https://assets.circle.so/kvx4ix1f5ctctk55daheobna46hf"
        kn.set_source_uri(uri)
        tbl = kn.load_source_canonical(file_type='pdf')
        # text = ('You took too long. You took too long. You are not easy to deal with. Payment Failure/Incorrect Payment. You provided '
        #         'me with incorrect information. Unhappy with delay. Unsuitable advice. You never answered my question. '
        #         'You did not understand my needs. I have been mis-sold. My details are not accurate. You have asked '
        #         'for too much information. You were not helpful. Payment not generated/received by customer. You did '
        #         'not keep me updated. Incorrect information given. The performance of my product was poor. No reply '
        #         'to customer contact. Requested documentation not issued. You did not explain the terms & conditions. '
        #         'Policy amendments not carried out. You did not explain the next steps/process to me. I cannot '
        #         'understand your letter/comms. Standard letter inappropriate. Customer payment processed incorrectly. '
        #         'All points not addressed. Could not understand the agent. Issue with terms and conditions. Misleading '
        #         'information. I can not use the customer portal. your customer portal is unhelpful')
        # arr = pa.array([text], pa.string())
        # tbl = pa.table([arr], names=['text'])
        sentences = tools.text_to_sentences(tbl)
        result = tools.text_to_chunks(sentences, char_chunk_size=500, overlap=10)
        print(kn.table_report(result).to_string())

    def test_text_chunk_semantic(self):
        kn = Knowledge.from_memory()
        tools: KnowledgeIntent = kn.tools
        # uri = "https://www.europarl.europa.eu/doceo/document/TA-9-2024-0138_EN.pdf"
        # kn.set_source_uri(uri)
        # tbl = kn.load_source_canonical(file_type='pdf')
        text = ('You took too long. You took too long. You are not easy to deal with. Payment Failure/Incorrect Payment. You provided '
                'me with incorrect information. Unhappy with delay. Unhappy with delay. Unsuitable advice. You never answered my question. '
                'You did not understand my needs. I have been mis-sold. My details are not accurate. You have asked '
                'for too much information. You were not helpful. Payment not generated/received by customer. You did '
                'not keep me updated. Incorrect information given. The performance of my product was poor. No reply '
                'to customer contact. Requested documentation not issued. You did not explain the terms & conditions. '
                'Policy amendments not carried out. You did not explain the next steps/process to me. I cannot '
                'understand your letter/comms. Standard letter inappropriate. Customer payment processed incorrectly. '
                'All points not addressed. Could not understand the agent. Issue with terms and conditions. Misleading '
                'information. I can not use the customer portal. your customer portal is unhelpful')
        arr = pa.array([text], pa.string())
        tbl = pa.table([arr], names=['text'])
        sentences = tools.text_to_sentences(tbl)
        result = tools.text_to_chunks(sentences, char_chunk_size=100)
        print(kn.table_report(result).to_string())

    def test_embedding(self):
        kn = Knowledge.from_memory()
        kn.set_source_uri("milvus://localhost:19530/rai")
        kn.set_persist_uri("milvus://localhost:19530/rai")
        tools: KnowledgeIntent = kn.tools
        text = ('You took too long. You are not easy to deal with. Payment Failure/Incorrect Payment. You provided '
                'me with incorrect information. Unhappy with delay. Unsuitable advice. You never answered my question. '
                'You did not understand my needs. I have been mis-sold. My details are not accurate. You have asked '
                'for too much information. You were not helpful. Payment not generated/received by customer. You did '
                'not keep me updated. Incorrect information given. The performance of my product was poor. No reply '
                'to customer contact. Requested documentation not issued. You did not explain the terms & conditions. '
                'Policy amendments not carried out. You did not explain the next steps/process to me. I cannot '
                'understand your letter/comms. Standard letter inappropriate. Customer payment processed incorrectly. '
                'All points not addressed. Could not understand the agent. Issue with terms and conditions. Misleading '
                'information. I can not use the customer portal. your customer portal is unhelpful')
        arr = pa.array([text], pa.string())
        tbl = pa.table([arr], names=['text'])
        sentences = tools.text_to_sentences(tbl)
        chunks = tools.text_to_chunks(sentences)
        # save
        kn.save_persist_canonical(chunks)
        result = kn.load_persist_canonical(query='long wait')
        print(kn.table_report(result).to_string())
        # kn.remove_canonical(kn.CONNECTOR_PERSIST)
        # result = kn.load_persist_canonical(query='long wait')
        # print(kn.table_report(result).to_string())

    def test_text_from_load(self):
        kn = Knowledge.from_memory()
        tools: KnowledgeIntent = kn.tools
        uri = '../../jupyter/knowledge/hadron/source/Global-Index-1st-Edition-Report.pdf'
        kn.set_source_uri(uri)
        tbl = kn.load_source_canonical(file_type='pdf', as_pages=True, as_markdown=True)
        result = tbl.filter(pc.greater(tbl['page_table_count'], 0))
        print(kn.table_report(result, headers='text', drop=True).to_string())

    def test_raise(self):
        startTime = datetime.now()
        with self.assertRaises(KeyError) as context:
            env = os.environ['NoEnvValueTest']
        self.assertTrue("'NoEnvValueTest'" in str(context.exception))
        print(f"Duration - {str(datetime.now() - startTime)}")


def get_table():
    n_legs = pa.array([2, 4, 5, 100])
    animals = pa.array(["Flamingo", "Horse", "Brittle stars", "Centipede"])
    names = ["n_legs", "animals"]
    return pa.Table.from_arrays([n_legs, animals], names=names)


def tprint(t: pa.table, top: int=None, headers: [str, list]=None, d_type: [str, list]=None, regex: [str, list]=None):
    top = top if isinstance(top, int) else 10
    _ = Commons.filter_columns(t.slice(0, top), headers=headers, d_types=d_type, regex=regex)
    print(Commons.table_report(_).to_string())


if __name__ == '__main__':
    unittest.main()



import os
from ds_core.handlers.abstract_handlers import AbstractSourceHandler, ConnectorContract
from ds_core.handlers.abstract_handlers import HandlerFactory, AbstractPersistHandler
from sentence_transformers import SentenceTransformer
import pyarrow as pa

__author__ = 'Darryl Oatridge'


class MilvusSourceHandler(AbstractSourceHandler):
    """ This handler class uses pymilvus package. Milvus is an open-source vector
    database designed to manage and search large-scale vector data. It is designed to
    store, index, and search vector data efficiently, targeting AI-driven applications
    that require efficient handling of unstructured data.

        URI example
            uri = "milvus://host:port/database?collection=<name>&reference=<name>"

        params:
            collection: The name of the collection
            reference: a prefix name to reference the document vector

        Environment:
            MILVUS_EMBEDDING_NAME
            MILVUS_EMBEDDING_DEVICE
            MILVUS_EMBEDDING_BATCH_SIZE
            MILVUS_EMBEDDING_DIM
            MILVUS_INDEX_CLUSTERS
            MILVUS_INDEX_SIMILARITY_TYPE
            MILVUS_QUERY_SEARCH_LIMIT
            MILVUS_QUERY_NUM_SIMILARITY
    """

    def __init__(self, connector_contract: ConnectorContract):
        """ initialise the Handler passing the Connector Contract """
        # required module import
        self.pymilvus = HandlerFactory.get_module('pymilvus')
        super().__init__(connector_contract)
        # reset to use dialect
        _kwargs = {**self.connector_contract.kwargs, **self.connector_contract.query}
        _database = self.connector_contract.path[1:] if self.connector_contract.path[1:] else 'rai'
        _embedding_name = os.environ.get('MILVUS_EMBEDDING_NAME', _kwargs.pop('embedding', 'all-mpnet-base-v2'))
        _device = os.environ.get('MILVUS_EMBEDDING_DEVICE', _kwargs.pop('device', 'cpu'))
        self._index_clusters = int(os.environ.get('MILVUS_INDEX_CLUSTERS', _kwargs.pop('index_clusters', '128')))
        self._index_type = os.environ.get('MILVUS_INDEX_SIMILARITY_TYPE', _kwargs.pop('index_type', 'L2'))
        self._search_limit = int(os.environ.get('MILVUS_QUERY_SEARCH_LIMIT', _kwargs.pop('search_limit', '8')))
        self._query_similarity = int(os.environ.get('MILVUS_QUERY_NUM_SIMILARITY', _kwargs.pop('query_similarity', '10')))
        self._batch_size = int(os.environ.get('MILVUS_EMBEDDING_BATCH_SIZE', _kwargs.pop('batch_size', '64')))
        self._dimensions = int(os.environ.get('MILVUS_EMBEDDING_DIM', _kwargs.pop('dim', '768')))
        self._reference = _kwargs.pop('reference', 'general')
        self._collection_name = _kwargs.pop('collection', "default")
        # embedding model
        self._embedding_model = SentenceTransformer(model_name_or_path=_embedding_name, device=_device)
        # Start the server
        self.pymilvus.connections.connect(host=connector_contract.hostname, port=connector_contract.port)
        if _database in self.pymilvus.db.list_database():
            self.pymilvus.db.using_database(_database)
        else:
            self.pymilvus.db.create_database(_database)
        # Create the collection
        if self.pymilvus.utility.has_collection(self._collection_name):
            self._collection = self.pymilvus.Collection(self._collection_name)
        else:
            fields = [
                self.pymilvus.FieldSchema(name="id", dtype=self.pymilvus.DataType.VARCHAR, auto_id=False, is_primary=True, max_length=64),
                self.pymilvus.FieldSchema(name="source", dtype=self.pymilvus.DataType.VARCHAR, max_length=1024),
                self.pymilvus.FieldSchema(name="embeddings", dtype=self.pymilvus.DataType.FLOAT_VECTOR, dim=self._dimensions)
            ]
            # schema
            schema = self.pymilvus.CollectionSchema(fields=fields)
            # collection
            self._collection = self.pymilvus.Collection(
                name=self._collection_name,
                schema=schema,
                num_shards=2,
                consistency_level='Strong')
            # index
            index_params = {
                "metric_type": self._index_type,
                "index_type": "IVF_FLAT",
                "params": {"nlist": self._index_clusters}
            }
            self._collection.create_index("embeddings", index_params)
        self._changed_flag = True

    def supported_types(self) -> list:
        """ The source types supported with this module"""
        return ['milvus']

    def exists(self) -> bool:
        """If the table exists"""
        return True

    def has_changed(self) -> bool:
        """ if the table has changed. Only works with certain implementations"""
        return self._changed_flag

    def reset_changed(self, changed: bool = False):
        """ manual reset to say the table has been seen. This is automatically called if the file is loaded"""
        changed = changed if isinstance(changed, bool) else False
        self._changed_flag = changed

    def load_canonical(self, query: [str, list], **kwargs) -> pa.Table:
        """ returns the canonical dataset based on a vector similarity search

            search(
                data: list[list[float]],
                anns_field: str,
                param: dict,
                limit: int
                expr: str | None,
                partition_names: list[str] | None,
                output_fields: list[str] | None,
                timeout: float | None,
                round_decimal: int
                )
)
        """
        if not isinstance(self.connector_contract, ConnectorContract):
            raise ValueError("The Connector Contract is not valid")
        expr = kwargs.get('expr', f"id like \"{str(self._reference)}_\"")
        # get collection
        if not self._collection:
            self._collection = self.pymilvus.Collection(self._collection_name)
        self._collection.load()
        # embedding
        query_vector = self._embedding_model.encode(query)
        # search
        params = {"metric_type": "L2", "params": {"nprobe": self._query_similarity}}
        results = self._collection.search(
            data = [query_vector],
            anns_field = "embeddings",
            param = params,
            expr = expr,
            limit = self._search_limit,
            output_fields=["source"])
        self._collection.release()
        # build table
        ids = pa.array(results[0].ids, pa.string())
        distances = pa.array(results[0].distances, pa.float32())
        entities = pa.array([x.entity.to_dict()['entity']['source'] for x in results[0]], pa.string())
        return pa.table([ids, distances, entities], names=['id', 'distance', 'source'])


class MilvusPersistHandler(MilvusSourceHandler, AbstractPersistHandler):
    # a Milvus persist handler

    def persist_canonical(self, canonical: pa.Table, **kwargs) -> bool:
        """ persists the canonical dataset"""
        return self.backup_canonical(canonical=canonical, **kwargs)

    def backup_canonical(self, canonical: pa.Table, **kwargs) -> bool:
        """ creates a backup of the canonical to an alternative table  """
        if not isinstance(self.connector_contract, ConnectorContract):
            return False
        _params = kwargs
        chunks = canonical.to_pylist()
        text_chunks = [item["chunk_text"] for item in chunks]
        embeddings = self._embedding_model.encode(text_chunks, batch_size=self._batch_size)
        data = [
            [f"{str(self._reference)}_{str(i)}" for i in range(len(text_chunks))],
            text_chunks,
            embeddings
        ]
        if not self._collection:
            self._collection = self.pymilvus.Collection(self._collection_name)
        self._collection.load()
        self._collection.upsert(data=data)
        self._collection.release()
        return

    def remove_canonical(self) -> bool:
        """removes a document reference"""
        if self.pymilvus.utility.has_collection(self._collection_name):
            if not self._collection:
                self._collection = self.pymilvus.Collection(self._collection_name)
            expr = f"id like \"{str(self._reference)}_\""
            self._collection.load()
            self._collection.delete(expr)
            self._collection.release()
        return True

    def remove_collection(self) -> bool:
        """remove a collection"""
        if self.pymilvus.utility.has_collection(self._collection_name):
            self.pymilvus.utility.drop_collection(self._collection_name)
        return True
